/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @file       mcv-test-pyr-fpx.cpp
 *
 * @brief      Tests the PYR Feature Extraction (FPX) process
 *
 * @author     Thomas Patton (thomas.patton@modalai.com)
 * @date       2024
 */

#include <fastcv.h>
#include <fcntl.h>
#include <getopt.h>
#include <modalcv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>   // for mkdir
#include <sys/types.h>  // for mode_t in mkdir
#include <time.h>
#include <unistd.h>

#include <iomanip>
#include <iostream>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#include <sstream>
#include "../src/cvp/internal/private_types.h"

#include "modalcv/io.hpp"
#include "modalcv/pyr.h"

// output files
#define OUTPUT_FOLDER "/data/libmodalcv/pyr_fpx/"

/**
 * @brief Helper fn for getting the current clock time in nanoseconds
 *
 * @return Time in ns
 */
static int64_t _apps_time_monotonic_ns() {
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
        fprintf(stderr, "ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec * 1000000000 + (int64_t)ts.tv_nsec;
}

int main(int argc, char* argv[]) {
    mcv_en_debug_prints(0);
    mcv_en_timing_prints(0);

    int w, h;
    std::vector<std::string> paths;
    if (init_modalcv_example(argc, argv, paths, w, h)) {
        fprintf(stderr, "ERROR, failed to initialize example\n");
        return -1;
    }

    std::string input_folder = argv[1];
    printf("%s\n", input_folder.c_str());

    // configure everything
    mcv_cvp_fpx_pyr_config_t pyr_fpx_config;
    pyr_fpx_config.width = w;
    pyr_fpx_config.height = h;
    pyr_fpx_config.n_octaves = 3;
    pyr_fpx_config.n_scales_per_octave = 1;
    pyr_fpx_config.mode = MCV_CVP_FPX_ZONE;
    pyr_fpx_config.nms_mode = MCV_CVP_FPX_5_TAP_NMS;
    pyr_fpx_config.score_threshold = 10;
    pyr_fpx_config.robustness = 10;  // 0-127, default 10
    pyr_fpx_config.zone_dims[0] = 16;
    pyr_fpx_config.zone_dims[1] = 8;
    pyr_fpx_config.zone_dims[2] = 4;
    // pyr_fpx_config.n_filter_coeffs[0] = 1;
    // pyr_fpx_config.n_filter_coeffs[1] = 2;
    // pyr_fpx_config.n_filter_coeffs[2] = 1;
    for (int i = 0; i < MCV_CVP_MAX_PYRAMID_OCTAVES; i++)
        pyr_fpx_config.enable_fpx[i] = 1;

    // memset(pyr_fpx_config.n_filter_coeffs, 3,
    //        MCV_CVP_MAX_PYRAMID_OCTAVES * sizeof(uint32_t));
    // memset(pyr_fpx_config.enable_fpx, 1, MCV_CVP_MAX_PYRAMID_OCTAVES);

    // create a pyr_fpx handle from our configuration
    mcv_cvp_pyr_fpx_handle handle = mcv_cvp_pyr_fpx_init(pyr_fpx_config);
    if (NULL == handle) {
        fprintf(stderr, "ERROR creating pyr fpx handle\n");
    }

    // create an output object
    mcv_cvp_pyr_fpx_t output_pyr_fpx;
    if (mcv_cvp_create_pyr_fpx(handle, &output_pyr_fpx)) {
        fprintf(stderr, "ERROR creating feature extraction pyramid\n");
        return -1;
    }

    // MAIN LOOP iterate over all images in folder
    printf("STARTING MAIN LOOP\n");
    for (size_t i = 0; i < paths.size(); i++) {
        std::stringstream ss;
        ss << std::setw(5) << std::setfill('0') << i;
        std::string input_path = paths[i];
        printf("%s\n", input_path.c_str());

        cv::Mat current_img_gray, current_img_color;
        current_img_gray = imread(input_path, cv::IMREAD_GRAYSCALE);
        // cv::blur(current_img_gray, current_img_gray, cv::Size(5, 5));
        cv::cvtColor(current_img_gray, current_img_color, cv::COLOR_GRAY2BGR);
        if (current_img_gray.type() != CV_8UC1) {
            fprintf(stderr, "ERROR loading images\n");
            return -1;
        }

        // run the pyr fpx process on the current frame
        int64_t start = _apps_time_monotonic_ns();
        mcv_cvp_pyr_fpx_process(handle, current_img_gray.data, &output_pyr_fpx);
        int64_t end = _apps_time_monotonic_ns();
        printf("Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);

        // draw features on the image
        for (size_t oct = 0; oct < pyr_fpx_config.n_octaves; oct++) {
            for (int i = 0; i < output_pyr_fpx.n_points[oct]; i++) {
                mcv_cvp_fpx_feature_t feature = output_pyr_fpx.fpx[oct][i];
                int feature_x = feature.x * pow(2, oct);  // upscale
                int feature_y = feature.y * pow(2, oct);  // upscale
                int radius = 4;
                int thickness = 1;

                cv::Scalar color;
                if (oct == 0) {
                    color = cv::Scalar(0, 255, 0);
                } else if (oct == 1) {
                    color = cv::Scalar(255, 0, 0);
                } else {
                    color = cv::Scalar(0, 0, 255);
                }

                cv::Rect rect(feature_x - radius, feature_y - radius,
                              radius * 2, radius * 2);
                cv::rectangle(current_img_color, rect, color, thickness, 1, 0);
            }
        }

        // write the score shift value
        pyr_fpx_session_t* session = (pyr_fpx_session_t*)handle;
        std::string score_shift_text = std::to_string(session->score_shift);

        // Set the position in the bottom-left corner
        int fontFace = cv::FONT_HERSHEY_SIMPLEX;
        double fontScale = 1.0;
        int thickness = 2;

        // Get the size of the text box
        int baseline = 0;
        cv::Size textSize = cv::getTextSize(score_shift_text, fontFace, fontScale, thickness, &baseline);
        
        // Position the text at the bottom-left corner (with some padding)
        cv::Point textOrg(10, current_img_color.rows - 10);  // (10, img.rows - 10) adds some padding from the bottom-left

        // Write the text onto the image
        cv::putText(current_img_color, score_shift_text, textOrg, fontFace, fontScale, cv::Scalar(255, 255, 255), thickness);

        std::string output_path = OUTPUT_FOLDER + ss.str() + "l.png";
        cv::imwrite(output_path, current_img_color);
    }

    // cleanup
    printf("all done, deinitializing\n");
    mcv_cvp_destroy_pyr_fpx(handle, &output_pyr_fpx);
    mcv_cvp_pyr_fpx_deinit(handle);
    printf("DONE\n");
    return 0;
}
